@@ -32,7 +32,7 @@
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  * OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * $Id: bims.c,v 1.13 2001/11/05 15:23:39 thhsieh Exp $
+ * $Id: bims.c,v 1.17 2002/04/04 16:28:23 thhsieh Exp $
  */
 #ifdef HAVE_CONFIG_H
 #include "../../../config.h"
@@ -55,6 +55,9 @@ struct _db_pool {
   int               len_pool;
 };
 
+struct TsiDB	*usertsidb;
+struct TsiYinDB	*useryindb;
+
 static int bimsTsiDBPoolSearch(struct _db_pool *_db, struct TsiInfo *ti);
 static int bimsTsiYinDBPoolSearch(struct _db_pool *_db, struct TsiYinInfo *ty);
 
@@ -68,6 +71,8 @@ static int  bimsVerifyPindown(struct bimsContext *bc,
 			      struct TsiYinInfo *ty, int yinoff, int index);
 
 static void bimsContextSmartEdit(struct _db_pool *_db, struct bimsContext *bc);
+static void bimsTsiyinDump(struct TsiDB *db, struct TsiYinDB *ydb);
+
 
 static ZuYinIndex bimsZozyKeyToZuYinIndex(int key);
 static ZuYinIndex bimsEtenKeyToZuYinIndex(int key);
@@ -130,6 +135,7 @@ bimsInit(char *tsidb_name, char *yindb_name)
   _db->tdb_pool = NULL;
   _db->ydb_pool = NULL;
   _db->len_pool = 0;
+
   return (DB_pool)_db;
 }
 
@@ -161,6 +167,79 @@ bimsDestroy(DB_pool db)
   free(_db);
 }
 
+
+int 
+bimsUserDBAppend(DB_pool db, char *usertsidb_name, char *useryindb_name)
+{
+    struct _db_pool *_db = (struct _db_pool *)db;
+    struct TsiDB *t;
+    struct TsiYinDB *y;
+    int len = 0;
+
+    /* check and open database */
+    if (! _db || !strlen(usertsidb_name) || !strlen(useryindb_name) )
+	return(-1);
+
+    t = tabeTsiDBOpen(DB_TYPE_DB, usertsidb_name,
+			DB_FLAG_OVERWRITE|DB_FLAG_CREATEDB|DB_FLAG_SHARED);
+    if (!t) {
+	return(-1);
+    }
+    usertsidb = t;
+
+    y = tabeTsiYinDBOpen(DB_TYPE_DB, useryindb_name,
+			DB_FLAG_OVERWRITE|DB_FLAG_CREATEDB|DB_FLAG_SHARED);
+    if (!y) {
+	t->Close(t);
+	return(-1);
+    }
+    useryindb = y;
+
+    /* append them to the pool */
+    if (_db->len_pool == 0) {
+	len = 2;  /* the primary plus this new one */
+	_db->tdb_pool = (struct TsiDB **)calloc(len, sizeof(struct TsiDB *));
+	_db->ydb_pool = (struct TsiYinDB **)calloc(len, sizeof(struct TsiYinDB *));
+
+	if (!_db->tdb_pool || !_db->ydb_pool) {
+	    t->Close(t);
+	    y->Close(y);
+	    return (-1);
+	}
+
+	/* put primary in */
+	_db->tdb_pool[0] = _db->tdb;
+	_db->ydb_pool[0] = _db->ydb;
+	/* append the new one */
+	_db->tdb_pool[1] = t;
+	_db->ydb_pool[1] = y;
+    }
+    else {
+	void *tmp;
+	len = _db->len_pool + 1;  /* the primary plus this new one */
+	tmp = (void *)realloc(_db->tdb_pool, len*sizeof(struct TsiDB *));
+	if (!tmp) {
+	    t->Close(t);
+	    y->Close(y);
+	    return (-1);
+	}
+	_db->tdb_pool = (struct TsiDB **)tmp;
+	tmp = (void *)realloc(_db->ydb_pool, len*sizeof(struct TsiYinDB *));
+	if (!tmp) {
+	    t->Close(t);
+	    y->Close(y);
+	    return (-1);
+	}
+	_db->ydb_pool = (struct TsiYinDB **)tmp;
+
+	/* append the new one */
+	_db->tdb_pool[_db->len_pool] = t;
+	_db->ydb_pool[_db->len_pool] = y;
+    }
+    _db->len_pool = len;
+    return (0);
+}
+
 /*
  * Append a pair of tsi and tsiyin db into the pool
  */
@@ -898,6 +977,7 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
   memset(yin, 0, sizeof(Yin)*TMP_BUFFER);
   memset(&tsi, 0, sizeof(tsi));
   memset(tmp, 0, sizeof(unsigned char)*TMP_BUFFER);
+  memset(&ty, 0, sizeof(ty));
   tsi.tsi = tmp;
 
   while (len > yinhead) {
@@ -921,9 +1001,13 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
       ysinfo[num_ysinfo].yindata = (Yin *)malloc(sizeof(Yin)*2);
       memcpy(ysinfo[num_ysinfo].yindata, bc->yin+yinhead, sizeof(Yin)*2);
       num_ysinfo++;
-      if(!bc->tsiboundary[yinhead+1]) {
+      if (!bc->tsiboundary[yinhead+1]) {
 	/* done duplicate a two-character word */
+/*
+	if (ty.tsidata)
+	  free(ty.tsidata);
 	memset(&ty, 0, sizeof(ty));
+*/
 	ty.yinlen = 2;
 	ty.yin = ysinfo[num_ysinfo-1].yindata;
 	rval = bimsTsiYinDBPoolSearch(_db, &ty);
@@ -954,7 +1038,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
       if (z != i) {
         continue;
       }
+/*
+      if (ty.tsidata)
+	free(ty.tsidata);
       memset(&ty, 0, sizeof(ty));
+*/
       ty.yinlen = i;
       memcpy(yin, bc->yin+yinhead, sizeof(Yin)*i);
       ty.yin = yin;
@@ -975,7 +1063,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
           if(z != j) {
             continue;
           }
+/*
+	  if (ty.tsidata)
+	    free(ty.tsidata);
 	  memset(&ty, 0, sizeof(ty));
+*/
 	  ty.yinlen = j;
 	  memcpy(yin, bc->yin+yinhead+i, sizeof(Yin)*j);
 	  ty.yin = yin;
@@ -1000,7 +1092,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
 	    if(z != k) {
 	      continue;
 	    }
+/*
+	    if (ty.tsidata)
+	      free(ty.tsidata);
 	    memset(&ty, 0, sizeof(ty));
+*/
 	    ty.yinlen = k;
 	    memcpy(yin, bc->yin+yinhead+i+j, sizeof(Yin)*k);
 	    ty.yin = yin;
@@ -1182,7 +1278,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
 	    comb[index].largest_sum = 0;
 	    k = (comb[index].s2-comb[index].s1);
 	    if (k > 1) {
+/*
+	      if (ty.tsidata);
+		free(ty.tsidata);
 	      memset(&ty, 0, sizeof(ty));
+*/
 	      ty.yinlen = k;
 	      memcpy(yin, bc->yin+comb[index].s1, sizeof(Yin)*k);
 	      ty.yin = yin;
@@ -1203,7 +1303,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
 	    }
 	    k = (comb[index].s3-comb[index].s2);
 	    if (k > 1) {
+/*
+	      if (ty.tsidata);
+		free(ty.tsidata);
 	      memset(&ty, 0, sizeof(ty));
+*/
 	      ty.yinlen = k;
 	      memcpy(yin, bc->yin+comb[index].s2, sizeof(Yin)*k);
 	      ty.yin = yin;
@@ -1224,7 +1328,11 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
 	    }
 	    k = (comb[index].len+comb[index].s1-comb[index].s3);
 	    if (k > 1) {
+/*
+	      if (ty.tsidata);
+		free(ty.tsidata);
 	      memset(&ty, 0, sizeof(ty));
+*/
 	      ty.yinlen = k;
 	      memcpy(yin, bc->yin+comb[index].s3, sizeof(Yin)*k);
 	      ty.yin = yin;
@@ -1307,6 +1415,8 @@ bimsContextDP(struct _db_pool *_db, struct bimsContext *bc)
     ncomb = 0;
     free(cand);
   }
+  if (ty.tsidata);
+    free(ty.tsidata);
 
   bc->ysinfo = ysinfo;
   bc->num_ysinfo = num_ysinfo;
@@ -1356,6 +1466,7 @@ bimsContextSmartEdit(struct _db_pool *_db, struct bimsContext *bc)
   memset(bc->internal_text, 0, sizeof(unsigned char)*(bc->yinlen*2+1));
   memset(&tsi, 0, sizeof(tsi));
   memset(tmp, 0, TMP_BUFFER);
+  memset(&ty, 0, sizeof(ty));
   tsi.tsi = tmp;
 
   for (i = 0; i < num_ysinfo; i++) {
@@ -1376,7 +1487,11 @@ bimsContextSmartEdit(struct _db_pool *_db, struct bimsContext *bc)
       }
     }
     else {
+/*
+      if (ty.tsidata)
+	free(ty.tsidata);
       memset(&ty, 0, sizeof(ty));
+*/
       ty.yinlen = ysinfo[i].yinlen;
       ty.yin = ysinfo[i].yindata;
       rval = bimsTsiYinDBPoolSearch(_db, &ty);
@@ -1905,6 +2020,153 @@ bimsSetMaxLen(unsigned long int bcid, int maxlen)
   return(0);
 }
 
+/*----------------------------------------------------------------------------
+
+	The tsiguess for bims, patched by
+
+		Pofeng Lee <informer@ns1.m2000.idv.tw>
+
+	edited by Tung-Han Hsieh <thhsieh@linux.org.tw>
+
+----------------------------------------------------------------------------*/
+/*
+ * copy from tsiguess.c 
+ */
+static int 
+isprep(unsigned char *zhi) 
+{
+/* 
+ * ref: http://www.dmpo.sinica.edu.tw:8000/~words/sou/sou.html
+ */
+  const char preplist[]= \
+/* s    */		"åBMPΩΦY" \
+/* Ƶ      */		"ܥu٬ҳȫh]nNN~" \
+/* ɺA    */		"F" \
+/* w/q */		"@GT|CKEQoӨӪMy" \
+/* yU    */		"ڰէaoaG" \
+/*       */		"bO" \
+/* m    */		"WUk" \
+/* Ϊʵ  */		"O@" \
+/* misc      */		"ڧApzLoe͢";
+
+  int i=0, len;
+
+  len = strlen(preplist);
+  for (i=0; i<len; i+=2) {
+    if (strncmp(zhi, preplist+i, 2) == 0)
+      return TRUE;
+  }
+  return FALSE;
+}
+
+/*
+ * copy from tsiguess.c
+ */
+#define MAX_REFCNT	1000
+static int
+tabe_guess_newtsi(struct ChunkInfo *chunk, struct TsiDB *newdb)
+{
+  int i, lbuf, buflen, need_update;
+  unsigned char  *tsi_str;
+  struct TsiInfo *tsi;		
+  char buf[1024];
+  Yin  yin[512], *pyin;
+
+  lbuf   = 1023;
+  pyin   = yin;
+  buf[0] = '\0';
+  need_update = 0;
+  for (i=0; i < chunk->num_tsi; i++) {
+    tsi_str = (chunk->tsi+i)->tsi ;
+    if (strlen(tsi_str) == 2 && isprep(tsi_str) == FALSE) {
+    /* r, [ buff */
+      strncat(buf, tsi_str, lbuf);
+      memmove(pyin, chunk->tsi[i].yindata, sizeof(Yin));
+      lbuf -= 2;
+      pyin ++;
+    }
+    else {
+    /* Or, }lz buf */
+      buflen = 1024 - lbuf - 1;
+      if (buflen >= 4) {  
+      /* osr */
+	tsi = tabeTsiInfoNew(buf);
+	newdb->Get(newdb, tsi);
+	if (tsi->yinnum == 0) {
+          tsi->yinnum = 1;
+          tsi->yindata = malloc(sizeof(Yin)*buflen/2);
+          memmove(tsi->yindata, yin, sizeof(Yin)*buflen/2);
+	  need_update = 1;
+        }
+	if (tsi->refcount < MAX_REFCNT)
+	  tsi->refcount ++;
+	newdb->Put(newdb, tsi);
+        if (need_update != 0) {
+          bimsTsiyinDump(usertsidb, useryindb);
+          need_update = 0;
+        }
+	tabeTsiInfoDestroy(tsi);
+      }
+      /* else r, if u, аO for spelling check */
+      buf[0] = '\0';
+      lbuf = 1023;
+      pyin = yin;
+    }
+  }
+
+  buflen = 1024 - lbuf - 1;
+  if (buflen >= 4) {
+  /* at the end of chunk, check again */
+    tsi = tabeTsiInfoNew(buf);
+    newdb->Get(newdb, tsi);
+    if (tsi->yinnum == 0) {
+      tsi->yinnum = 1;
+      tsi->yindata = malloc(sizeof(Yin)*buflen/2);
+      memmove(tsi->yindata, yin, sizeof(Yin)*buflen/2);
+      need_update = 1;
+    }
+    if (tsi->refcount < MAX_REFCNT)
+      tsi->refcount ++;
+    newdb->Put(newdb,tsi);
+    if (need_update != 0) {
+      bimsTsiyinDump(usertsidb, useryindb);
+      need_update = 0;
+    }
+    tabeTsiInfoDestroy(tsi);
+  }
+  return TRUE;
+}
+
+/* 
+ * Guess Tsi from a string 
+ */
+static void
+bimsTsiGuess(struct TsiDB *tdb, struct TsiDB *usertsidb, char *str)
+{
+  int    i=0;
+  struct ChuInfo *chu=NULL;
+
+  chu = (struct ChuInfo *) malloc(sizeof(struct ChuInfo));
+  chu->chu = (char *) malloc(sizeof(unsigned char)*strlen(str)+1);
+  chu->num_chunk = 0;
+  chu->chunk = NULL;
+  strcpy(chu->chu, str);
+
+  tabeChuInfoToChunkInfo(chu);
+  for (i=0; i < chu->num_chunk; i++) {
+    tabeChunkSegmentationComplex(tdb, chu->chunk+i);
+    tabe_guess_newtsi(chu->chunk+i, usertsidb);
+  }
+  tabeChunkInfoDestroy(chu->chunk);
+  free(chu->chu);
+  free(chu);
+}
+/*----------------------------------------------------------------------------
+
+	End of tsiguess.
+
+----------------------------------------------------------------------------*/
+
 /*
  * fetch some text from head
  */
@@ -1920,6 +2182,8 @@ bimsFetchText(DB_pool db, unsigned long int bcid, int len)
 
   bc = bimsGetBC(bcid);
   newlen = (bc->yinlen < len) ? bc->yinlen : len;
+  memset(&ti, 0, sizeof(ti));
+  memset(&yi, 0, sizeof(yi));
 
   /* see if we need to/can update DBs */
   if (bc->updatedb &&
@@ -1928,9 +2192,10 @@ bimsFetchText(DB_pool db, unsigned long int bcid, int len)
     int ylen, yoff;
 
     for (i = 0; i < bc->num_ysinfo; i++) {
+/*
       memset(&ti, 0, sizeof(ti));
       memset(&yi, 0, sizeof(yi));
-
+*/
       ylen = bc->ysinfo[i].yinlen;
       yoff = bc->ysinfo[i].yinoff;
       /*
@@ -2027,6 +2292,7 @@ bimsFetchText(DB_pool db, unsigned long int bcid, int len)
   bc->yinlen -= newlen;
 
   bimsContextSmartEdit(_db, bc);
+  bimsTsiGuess(_db->tdb, usertsidb, str);
 
   return(str);
 }
@@ -2886,6 +3152,50 @@ bimsHsuZuYinContextInput(struct ZuYinContext *zc, int index)
   return(rval);
 }
 
+/*
+ * Check whether the specified tsi exists and add it to the user's database.
+ */
+
+void         
+bimsUserTsiEval(DB_pool db, struct TsiInfo *tsi, struct TsiYinInfo *ty)
+{
+    struct _db_pool *_db = (struct _db_pool *)db;
+    int rval;
+    int i;
+    int found=0;
+
+    /* Lookup the user-specified Tsi in the DB pool */
+    rval = bimsTsiYinDBPoolSearch(_db, ty); 
+
+    /* Check for Tsi which has identical Yin and Zhi */
+    for(i=0; i < ty->tsinum; i++){
+	if(strncmp(ty->tsidata+i*ty->yinlen*2, tsi->tsi, ty->yinlen*2)==0){
+	    found=1; 
+	    break;
+	}
+    }
+    if (found)
+	return;
+    else {
+	int r;
+
+	tsi->refcount++;
+	tsi->yinnum++;
+
+	/* 32 should be more than adequate */
+	tsi->yindata = (Yin*)realloc(tsi->yindata, sizeof(Yin)*32);
+
+	memcpy(tsi->yindata+(tsi->yinnum-1)*ty->yinlen, ty->yin,
+		sizeof(Yin)*ty->yinlen);
+
+	/* Add a new entry in user's database */
+	r = usertsidb->Put(usertsidb, tsi);
+
+	/* Update user's TsiYinDB */
+	bimsTsiyinDump(usertsidb, useryindb);
+    }
+}
+
 /*
  *  bims to tabe operation interfaces.
  */
@@ -2915,3 +3225,84 @@ bimstabeZhiToYin(DB_pool db, struct TsiInfo *zhi)
   }
   return zhuyin;
 }
+
+
+/*
+ *  Stolen from tsiyindump.c from utils for dumping user's Tsi to
+ *  user's Yin DB.
+ */
+static void
+bimsTsiyinDump(struct TsiDB *db, struct TsiYinDB *ydb)
+{
+    struct TsiInfo tsi;
+    struct TsiYinInfo tsiyin;
+    unsigned char str[80];
+    int rval, i, j, len;
+
+    rval = db->RecordNumber(db);
+    if (rval < 0) {
+	fprintf(stderr, "bimsTsiyinDump: wrong DB format.\n");
+    }
+
+    tsi.tsi = (ZhiStr)str;
+    memset(tsi.tsi, 0, 80);
+    tsi.refcount = -1;
+    tsi.yinnum = -1;
+    tsi.yindata = (Yin *)NULL;
+    memset(&tsiyin, 0, sizeof(struct TsiYinInfo));
+
+    i = 0;
+    while (1) {
+	if (i == 0)
+	    db->CursorSet(db, &tsi, 0);
+	else {
+	    rval = db->CursorNext(db, &tsi);
+	    if (rval < 0) 
+		break;
+	}
+	i++;
+
+	if (!tsi.yinnum) {
+	    tabeTsiInfoLookupPossibleTsiYin(db, &tsi);
+	}
+	len = strlen((char *)tsi.tsi)/2;
+	for (j=0; j < tsi.yinnum; j++) {
+	    tsiyin.yinlen = len;
+	    tsiyin.yin = (Yin *)malloc(sizeof(Yin)*len);
+	    memcpy(tsiyin.yin, tsi.yindata+j*len, sizeof(Yin)*len);
+	    rval = ydb->Get(ydb, &tsiyin);
+	    if (rval < 0) { /* no such tsiyin */
+		tsiyin.tsinum = 1;
+		tsiyin.tsidata = (ZhiStr)malloc(sizeof(unsigned char)*len*2);
+		memcpy(tsiyin.tsidata, tsi.tsi, sizeof(unsigned char)*len*2);
+		ydb->Put(ydb, &tsiyin);
+	    }
+	    else {
+		int k, found=0;
+		for (k=0; k<tsiyin.tsinum; k++) {
+		    if (strncmp(tsiyin.tsidata+k*len*2, tsi.tsi, len*2)==0) {
+			found = 1;
+			break;
+		    }
+		}
+		if (found == 0) {
+		    tsiyin.tsidata = (ZhiStr)realloc(tsiyin.tsidata,
+			sizeof(unsigned char)*((tsiyin.tsinum+1)*len*2));
+		    memcpy(tsiyin.tsidata+(tsiyin.tsinum*len*2), tsi.tsi,
+			sizeof(unsigned char)*len*2);
+		    tsiyin.tsinum++;
+		    ydb->Put(ydb, &tsiyin);
+		}
+	    }
+	    free(tsiyin.yin);
+	    if (tsiyin.tsidata) {
+		free(tsiyin.tsidata);
+		tsiyin.tsidata = NULL;
+	    }
+	}
+	if (tsi.yindata) {
+	    free(tsi.yindata);
+	    tsi.yindata = NULL;
+	}
+    }
+}
